"Utilites for Cognite API SDK\n\nThis module provides helper methods and different utilities for the Cognite API Python SDK.\n\nThis module is protected and should not used by end-users.\n"
import gzip
import json
import logging
import re
import time
from datetime import datetime, timezone
import requests
import cognite.config as config

log = logging.getLogger("cognite-sdk")


def delete_request(url, params=None, headers=None, cookies=None):
    "Perform a DELETE request with a predetermined number of retries."
    _log_request("DELETE", url, params=params, headers=headers, cookies=cookies)
    for number_of_tries in range((config.get_number_of_retries() + 1)):
        try:
            res = requests.delete(url, params=params, headers=headers, cookies=cookies)
            if res.status_code == 200:
                return res
        except Exception:
            if number_of_tries == config.get_number_of_retries():
                raise
    x_request_id = res.headers.get("X-Request-Id")
    code = res.status_code
    try:
        error = res.json()["error"]
        if isinstance(error, str):
            msg = error
        else:
            msg = error.get("message") or res.json()
    except TypeError:
        msg = res.content
    except KeyError:
        msg = res.json()
    raise APIError(msg, code, x_request_id)


def get_request(url, params=None, headers=None, cookies=None):
    "Perform a GET request with a predetermined number of retries."
    _log_request("GET", url, params=params, headers=headers, cookies=cookies)
    for number_of_tries in range((config.get_number_of_retries() + 1)):
        try:
            res = requests.get(url, params=params, headers=headers, cookies=cookies)
            if res.status_code == 200:
                return res
        except Exception:
            if number_of_tries == config.get_number_of_retries():
                raise
    x_request_id = res.headers.get("X-Request-Id")
    code = res.status_code
    try:
        error = res.json()["error"]
        if isinstance(error, str):
            msg = error
        else:
            msg = error.get("message") or res.json()
    except TypeError:
        msg = res.content
    except KeyError:
        msg = res.json()
    raise APIError(msg, code, x_request_id)


def post_request(url, body, headers=None, params=None, cookies=None, use_gzip=False):
    "Perform a POST request with a predetermined number of retries."
    _log_request("POST", url, body=body, params=params, headers=headers, cookies=cookies)
    for number_of_tries in range((config.get_number_of_retries() + 1)):
        try:
            if use_gzip:
                if headers:
                    headers["Content-Encoding"] = "gzip"
                else:
                    headers = {"Content-Encoding": "gzip"}
                res = requests.post(
                    url,
                    data=gzip.compress(json.dumps(body).encode("utf-8")),
                    headers=headers,
                    params=params,
                    cookies=cookies,
                )
            else:
                res = requests.post(url, data=json.dumps(body), headers=headers, params=params, cookies=cookies)
            if res.status_code == 200:
                return res
        except Exception:
            if number_of_tries == config.get_number_of_retries():
                raise
    x_request_id = res.headers.get("X-Request-Id")
    code = res.status_code
    try:
        error = res.json()["error"]
        if isinstance(error, str):
            msg = error
        else:
            msg = error.get("message") or res.json()
    except TypeError:
        msg = res.content
    except KeyError:
        msg = res.json()
    raise APIError(msg, code, x_request_id)


def put_request(url, body=None, headers=None, cookies=None):
    "Perform a PUT request with a predetermined number of retries."
    _log_request("PUT", url, body=body, headers=headers, cookies=cookies)
    for number_of_tries in range((config.get_number_of_retries() + 1)):
        try:
            res = requests.put(url, data=json.dumps(body), headers=headers, cookies=cookies)
            if res.ok:
                return res
        except Exception:
            if number_of_tries == config.get_number_of_retries():
                raise
    x_request_id = res.headers.get("X-Request-Id")
    code = res.status_code
    try:
        error = res.json()["error"]
        if isinstance(error, str):
            msg = error
        else:
            msg = error.get("message") or res.json()
    except TypeError:
        msg = res.content
    except KeyError:
        msg = res.json()
    raise APIError(msg, code, x_request_id)


def _log_request(method, url, **kwargs):
    log.debug("HTTP/1.1 {} {}".format(method, url), extra=kwargs)


def datetime_to_ms(dt):
    return int((dt.replace(tzinfo=timezone.utc).timestamp() * 1000))


def round_to_nearest(x, base):
    return int((base * round((float(x) / base))))


def granularity_to_ms(time_string):
    "Returns millisecond representation of granularity time string"
    magnitude = int("".join([c for c in time_string if c.isdigit()]))
    unit = "".join([c for c in time_string if c.isalpha()])
    unit_in_ms = {
        "s": 1000,
        "second": 1000,
        "m": 60000,
        "minute": 60000,
        "h": 3600000,
        "hour": 3600000,
        "d": 86400000,
        "day": 86400000,
    }
    return magnitude * unit_in_ms[unit]


def get_aggregate_func_return_name(agg_func):
    agg_funcs = {
        "avg": "average",
        "average": "average",
        "count": "count",
        "continuousvariance": "continuousvariance",
        "cv": "continuousvariance",
        "discretevariance": "discretevariance",
        "dv": "discretevariance",
        "int": "interpolation",
        "interpolation": "interpolation",
        "max": "max",
        "min": "min",
        "step": "stepinterpolation",
        "stepinterpolation": "stepinterpolation",
        "sum": "sum",
        "totalvariation": "totalvariation",
        "tv": "totalvariation",
    }
    return agg_funcs.get(agg_func)


def _time_ago_to_ms(time_ago_string):
    "Returns millisecond representation of time-ago string"
    if time_ago_string == "now":
        return 0
    pattern = "(\\d+)([a-z])-ago"
    res = re.match(pattern, str(time_ago_string))
    if res:
        magnitude = int(res.group(1))
        unit = res.group(2)
        unit_in_ms = {"s": 1000, "m": 60000, "h": 3600000, "d": 86400000, "w": 604800000}
        return magnitude * unit_in_ms[unit]
    return None


def interval_to_ms(start, end):
    "Returns the ms representation of start-end-interval whether it is time-ago, datetime or None."
    time_now = int(round((time.time() * 1000)))
    if isinstance(start, datetime):
        start = datetime_to_ms(start)
    elif isinstance(start, str):
        start = time_now - _time_ago_to_ms(start)
    elif start is None:
        start = time_now - _time_ago_to_ms("2w-ago")
    if isinstance(end, datetime):
        end = datetime_to_ms(end)
    elif isinstance(end, str):
        end = time_now - _time_ago_to_ms(end)
    elif end is None:
        end = time_now
    return (start, end)


class APIError(Exception):
    def __init__(self, message, code=None, x_request_id=None):
        self.message = message
        self.code = code
        self.x_request_id = x_request_id

    def __str__(self):
        return "{} | code: {} | X-Request-ID: {}".format(self.message, self.code, self.x_request_id)


class InputError(Exception):
    pass


class Bin:
    "\n    Attributes:\n        entries (List): List of entries.\n        get_count (Callable): Callable function to get count.\n    "

    def __init__(self, get_count):
        "\n        Args:\n            get_count: A function that will take an element and get the count of something in it.\n        "
        self.entries = []
        self.get_count = get_count

    def add_item(self, item):
        self.entries.append(item)

    def sum(self):
        total = 0
        for elem in self.entries:
            total += self.get_count(elem)
        return total

    def show(self):
        return self.entries


def first_fit(list_items, max_size, get_count):
    "Returns list of bins with input items inside."
    list_items = sorted(list_items, key=get_count, reverse=True)
    list_bins = [Bin(get_count=get_count)]
    for item in list_items:
        alloc_flag = False
        for bin in list_bins:
            if (bin.sum() + get_count(item)) <= max_size:
                bin.add_item(item)
                alloc_flag = True
                break
        if not alloc_flag:
            new_bin = Bin(get_count=get_count)
            new_bin.add_item(item)
            list_bins.append(new_bin)
    list_items = []
    for bin in list_bins:
        list_items.append(bin.show())
    return list_items


def to_camel_case(snake_case_string):
    components = snake_case_string.split("_")
    return components[0] + "".join((x.title() for x in components[1:]))


def to_snake_case(camel_case_string):
    s1 = re.sub("(.)([A-Z][a-z]+)", "\\1_\\2", camel_case_string)
    return re.sub("([a-z0-9])([A-Z])", "\\1_\\2", s1).lower()



"Timeseries Module\n\nThis module mirrors the Timeseries API. It allows you to fetch data from the api and output it in various formats.\n\nhttps://doc.cognitedata.com/0.6/#Cognite-API-Time-series\n"
from cognite import _utils, config
from cognite.v05.dto import TimeSeriesResponse


def get_time_series_by_id(id, include_metadata=False, **kwargs):
    "Returns a TimeseriesResponse object containing the requested timeseries.\n\n    Args:\n        id (int):           ID of timeseries to look up\n\n        include_metadata (bool):    Decide if the metadata field should be returned or not. Defaults to False.\n\n    Keyword Arguments:\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v05.dto.TimeSeriesResponse: A data object containing the requested timeseries.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/timeseries/{}".format(project, id)
    headers = {"api-key": api_key, "accept": "application/json"}
    params = {"includeMetadata": include_metadata}
    res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
    return TimeSeriesResponse(res.json())


def get_multiple_time_series_by_id(ids, include_metadata=False, **kwargs):
    "Returns a TimeseriesResponse object containing the requested timeseries.\n\n    Args:\n        ids (List[int]):           IDs of timeseries to look up\n\n    Keyword Arguments:\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v05.dto.TimeSeriesResponse: A data object containing the requested timeseries with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/timeseries/byids".format(project)
    headers = {"api-key": api_key, "accept": "application/json", "content-type": "application/json"}
    body = {"items": ids}
    params = {"includeMetadata": include_metadata}
    res = _utils.post_request(url=url, body=body, params=params, headers=headers, cookies=config.get_cookies())
    return TimeSeriesResponse(res.json())



"Preprocessing module.\n\nThis module provides a number of preprocessing methods to further clean the data retrieved from the Cognite API.\n"
import math
import warnings
from functools import reduce
import numpy as np
import pandas as pd


def _merge_dataframes(df1, df2):
    if df1 is None:
        return df2
    if df2 is None:
        return df1
    return pd.merge(df1, df2, on="timestamp", how="outer").sort_values(by="timestamp").reset_index(drop=True)


def merge_list_of_dataframes(dataframes):
    "Merges together a list of dataframes and creates a time index with evenly spaced intervals.\n\n    Adds NaN to timestamps with missing values.\n\n    Args:\n        dataframes (list(pandas.DataFrame)):  Input dataframes.\n\n    Returns:\n        pandas.DataFrame: Dataframes merged into one with evenly spaced intervals.\n    "
    df_merged = reduce((lambda x, y: _merge_dataframes(x, y)), dataframes)
    df_even_index = make_index_even(df_merged)
    return df_even_index


def make_index_even(dataframe):
    "Creates time index with evenly spaced intervals. Adds NaN to timestamps with missing values.\n\n    Args:\n        dataframe (pandas.DataFrame):  Input dataframe.\n\n    Returns:\n        pandas.DataFrame: Input dataframe with evenly spaced intervals.\n    "
    timestamps = dataframe.timestamp.values
    start_time = timestamps[0]
    end_time = timestamps[(-1)]
    deltas = np.diff(timestamps, 1)
    delta = reduce((lambda x, y: math.gcd(x, y)), deltas)
    t_new = np.arange(start_time, (end_time + delta), delta)
    new_df = pd.DataFrame(t_new, columns=["timestamp"])
    return new_df.merge(dataframe, on="timestamp", how="outer").sort_values(by="timestamp").reset_index(drop=True)


def fill_nan(dataframe):
    "Uses step interpolation to replace NaN values with the previous non-NaN value.\n\n    Args:\n        dataframe (pandas.DataFrame):  Input dataframe.\n\n    Returns:\n        pandas.DataFrame: Input dataframe with NaN values removed by forward fill.\n    "
    return dataframe.fillna(method="ffill")


def remove_nan_columns(dataframe):
    "Removes columns of data frame where any value is NaN.\n\n    Args:\n        dataframe (pandas.DataFrame):  Input dataframe.\n\n    Returns:\n        tuple: tuple containing:\n            pandas.DataFrame: Dataframe with columns containing NaN values removed.\n            numpy.array: Array of bools indicating which columns were kept.\n    "
    df_copy = dataframe.set_index("timestamp")
    selected_columns_mask = df_copy.notnull().all().values
    return (dataframe.dropna(axis=1, how="any"), selected_columns_mask)


def normalize(dataframe):
    "Centers and scales each column in the data frame to zero mean and unit variance.\n\n    Args:\n        dataframe (pandas.DataFrame):  Input dataframe.\n\n    Returns:\n        pandas.DataFrame: Normalized dataframe.\n    "
    dataframe = dataframe.set_index("timestamp")
    dataframe = (dataframe - dataframe.mean()) / dataframe.std()
    return dataframe.reset_index()


def remove_zero_variance_columns(dataframe):
    "Removes columns with zero variance.\n\n    Args:\n        dataframe (pandas.DataFrame):  Input dataframe.\n\n    Returns:\n        tuple: tuple containing:\n            pandas.DataFrame: Dataframe with zero-variance columns removed.\n            numpy.array: Array of bools indicating which columns were kept.\n    "
    dataframe = dataframe.set_index("timestamp")
    selected_columns_mask = dataframe.var().values > 0
    dataframe = dataframe.loc[:, selected_columns_mask].reset_index()
    return (dataframe, selected_columns_mask)


def preprocess(dataframe, remove_leading_nan_rows=False, center_and_scale=False):
    "Performs a series of preprocessing steps on the given dataframe.\n\n    1) Creates an evenly spaced time index\n    2) Forward fills NaN values\n    3) Either removes leading rows with nan values or removes columns with leading nan values\n    4) Removes columns with zero variance\n    5) Optionally centers and scales the dataframe to zero mean and unit variance.\n\n    Args:\n        dataframe (pandas.DataFrame):  Input dataframe.\n        remove_leading_nan_rows (bool, optional): Whether or not to skip leading rows containing NaN values.\n        center_and_scale (bool, optional): Whether or not to normalize the data.\n\n    Returns:\n        tuple: tuple containing:\n            pandas.DataFrame: Dataframe with zero-variance columns removed.\n            numpy.array: Array of bools indicating which columns were kept.\n    "
    dataframe = make_index_even(dataframe)
    dataframe = fill_nan(dataframe)
    if remove_leading_nan_rows:
        dataframe.dropna(axis=0, how="any", inplace=True)
    if dataframe.isnull().sum().values.sum():
        warning = "Some columns contain leading NaN-values after forward-fill. Columns {} will be removed.\nConsider setting the skip_leading_nan_rows parameter to True in order to remove leading rows with NaN values.".format(
            dataframe.columns[dataframe.isnull().any().values].values
        )
        warnings.warn(warning)
    (_, selected_columns_mask_nan) = remove_nan_columns(dataframe)
    (_, selected_columns_mask_zero_var) = remove_zero_variance_columns(dataframe)
    selected_columns_mask = np.logical_and(selected_columns_mask_nan, selected_columns_mask_zero_var)
    dataframe = dataframe.set_index("timestamp").loc[:, selected_columns_mask].reset_index()
    if center_and_scale:
        dataframe = normalize(dataframe)
    return (dataframe, selected_columns_mask)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import argparse
import os
import sys

from cognite.cli.cli_models import CogniteModelsCLI


class CogniteCLI:
    def __init__(self):
        """cognite <service> [<args>]

The available Cognite services are:
    models      Model hosting
"""
        parser = argparse.ArgumentParser(description="Command line interface for CDP.", usage=self.__init__.__doc__)
        parser.add_argument("service", help="Service to run")
        # parse_args defaults to [1:] for args, but you need to
        # exclude the rest of the args too, or validation will fail
        args = parser.parse_args(sys.argv[1:2])
        if not os.getenv("COGNITE_API_KEY") or not os.getenv("COGNITE_PROJECT"):
            print("You must set COGNITE_API_KEY and COGNITE_PROJECT environment variables.")
            return
        if not hasattr(self, args.service.replace("-", "_")):
            print("Unrecognized service")
            parser.print_help()
            return
        # use dispatch pattern to invoke method with same name
        getattr(self, args.service)(sys.argv[2:])

    def models(self, args):
        """cognite models <command> [<args>]

The available commands are:
    source      Create a new model source package in the current directory.
    deploy          Run tests, upload a source package and optionally create a model with a specified name.
    get         Get all models, source packages, or versions for a model.
"""
        parser = argparse.ArgumentParser(description="Access CDP model hosting services.", usage=self.models.__doc__)
        parser.add_argument("command", help="Command to run", choices=["source", "get", "deploy"])
        parsed_args = parser.parse_args(args[:1])
        ml_cli = CogniteModelsCLI()
        getattr(ml_cli, parsed_args.command)(args[1:])


def main():
    CogniteCLI()



import json
from datetime import datetime
from io import BytesIO
import pandas as pd
from cognite import config
from cognite._utils import InputError, get_aggregate_func_return_name, to_camel_case, to_snake_case
from cognite.v05 import files
from cognite.v05 import timeseries as time_series_v05
from cognite.v06 import time_series as time_series_v06


class TimeSeries:
    "Object for specifying a specific time series from the TimeSeries API when using a data spec.\n\n    Args:\n        id (int): id of time series to retrieve\n        aggregates (List[str]): Local aggregate functions to apply\n        missing_data_strategy(str): Missing data strategy to apply\n        label (str): name of the column in the resulting data frame when passed to data transfer service.\n\n    Examples:\n        When you supply a label the resulting data frames produced by data transfer service\n        will use the label as column names.\n    "

    def __init__(self, id, aggregates=None, missing_data_strategy=None, label=None):
        self.id = id
        self.aggregates = aggregates
        self.missing_data_strategy = missing_data_strategy
        self.label = label
        self._name = None


class TimeSeriesDataSpec:
    'Object for specifying data from the TimeSeries API when using a data spec.\n\n    Args:\n        time_series (List[data_transfer_service.TimeSeries]):  Time series\n        aggregates (List[str]): List of aggregate functions\n        granularity (str): Granularity of aggregates\n        missing_data_strategy (str): Missing data strategy to apply, can be "linearInterpolation" or "ffill"\n        start (Union[str, int, datetime]): Start time\n        end (Union[str, int, datetime]): end time\n        label (str): Label for this data spec\n\n    Examples:\n        When you specify a label, data transfer service will exhibit the following behaviour::\n\n            ts_data_spec = TimeSeriesDataSpec(..., label="some_label")\n            data_spec = DataSpec([ts_data_spec], ...)\n            dts = DataTransferService(data_spec)\n            dataframes = dts.get_dataframes\n            my_df = dataframes["some_label"]\n    '

    def __init__(
        self, time_series, aggregates, granularity, missing_data_strategy=None, start=None, end=None, label=None
    ):
        self.time_series = time_series
        self.aggregates = aggregates
        self.granularity = granularity
        self.missing_data_strategy = missing_data_strategy
        self.start = start
        self.end = end
        self.label = label or "default"


class FilesDataSpec:
    "Object for specifying data from the Files API when using a data spec.\n\n    Args:\n        file_ids (Dict[str, int]):    Dictionary of fileNames -> fileIds\n    "

    def __init__(self, file_ids):
        self.file_ids = file_ids


class DataSpec:
    "Object for specifying data when querying CDP.\n\n    Args:\n        time_series_data_specs (List[TimeSeriesDataSpec]):  Time Series data specs\n        files_data_spec (FilesDataSpec):                    Files data spec\n\n    Raises:\n        DataSpecValidationError: An error occurred while validating the data spec\n\n    "

    def __init__(self, time_series_data_specs=None, files_data_spec=None):
        self.time_series_data_specs = time_series_data_specs
        self.files_data_spec = files_data_spec
        self.__validate_time_series_data_specs()
        self.__validate_files_data_spec()

    def __validate_time_series_data_specs(self):
        if self.time_series_data_specs:
            labels = []
            for tsds in self.time_series_data_specs:
                if not isinstance(tsds, TimeSeriesDataSpec):
                    raise DataSpecValidationError("Time series data specs must be TimeSeriesDataSpec objects.")
                label = tsds.label or "default"
                if label in labels:
                    raise DataSpecValidationError("Unique labels for each dataspec must be used")
                labels.append(label)
                if not isinstance(tsds.time_series, list):
                    raise DataSpecValidationError("Time series must be a list of time series objects.")
                if not (len(tsds.time_series) > 0):
                    raise DataSpecValidationError("A time series data spec does not contain any time series")
                ts_labels = []
                for ts in tsds.time_series:
                    if not isinstance(ts, TimeSeries):
                        raise DataSpecValidationError("Time series must be a TimeSeries object")
                    if ts.label and (ts.label in ts_labels):
                        raise DataSpecValidationError("Time series labels must be unique")
                    ts_labels.append(ts.label)

    def __validate_files_data_spec(self):
        if self.files_data_spec:
            if not isinstance(self.files_data_spec, FilesDataSpec):
                raise DataSpecValidationError("files_data_spec must be a FilesDataSpec object")
            if not isinstance(self.files_data_spec.file_ids, dict):
                raise DataSpecValidationError("file_ids must be a dict of form {name: id}")
            for (name, id) in self.files_data_spec.file_ids.items():
                if not isinstance(name, str):
                    raise DataSpecValidationError("File names must be a strings")
                if not isinstance(id, int):
                    raise DataSpecValidationError("File ids must be integers")

    def to_JSON(self):
        return json.dumps(
            {to_camel_case(key): value for (key, value) in self.__dict__.items()}, cls=self.__DataSpecEncoder
        )

    @classmethod
    def from_JSON(cls, json_repr):
        json_repr = json.loads(json_repr)
        time_series_data_specs_json = json_repr.get("timeSeriesDataSpecs")
        files_data_spec_json = json_repr.get("filesDataSpec")
        if not (time_series_data_specs_json or files_data_spec_json):
            raise DataSpecValidationError("Not a valid data spec")
        time_series_data_specs = []
        if time_series_data_specs_json:
            for tsds_json in time_series_data_specs_json:
                time_series_json = tsds_json.get("timeSeries")
                if time_series_json:
                    tsds = TimeSeriesDataSpec(**{to_snake_case(key): value for (key, value) in tsds_json.items()})
                    tsds.time_series = [
                        TimeSeries(**{to_snake_case(key): value for (key, value) in ts.items()})
                        for ts in time_series_json
                    ]
                    time_series_data_specs.append(tsds)
        files_data_spec = None
        if files_data_spec_json:
            files_data_spec = FilesDataSpec(
                **{to_snake_case(key): value for (key, value) in files_data_spec_json.items()}
            )
        return DataSpec(time_series_data_specs, files_data_spec)

    class __DataSpecEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, (TimeSeries, TimeSeriesDataSpec, FilesDataSpec)):
                new_dict = {}
                for (key, value) in obj.__dict__.items():
                    if value is not None:
                        new_dict[to_camel_case(key)] = value
                return new_dict
            elif obj is None:
                del obj
            return super(self, self).default(obj)


class DataSpecValidationError(Exception):
    pass


class DataTransferService:
    "Create a Data Transfer Service object.\n\n    Fetch timeseries from the api.\n    "

    def __init__(self, data_spec, project=None, api_key=None, cookies=None, num_of_processes=10):
        "\n        Args:\n            data_spec (data_transfer_service.DataSpec):   Data Spec.\n            project (str):          Project name.\n            api_key (str):          Api key.\n            cookies (dict):         Cookies.\n        "
        (config_api_key, config_project) = config.get_config_variables(api_key, project)
        if isinstance(data_spec, DataSpec):
            self.data_spec = data_spec
        elif isinstance(data_spec, str):
            self.data_spec = DataSpec.from_JSON(data_spec)
        else:
            raise InputError("DataTransferService accepts a DataSpec instance or a json represantion of it.")
        self.ts_data_specs = self.data_spec.time_series_data_specs
        self.files_data_spec = self.data_spec.files_data_spec
        self.api_key = api_key or config_api_key
        self.project = project or config_project
        self.cookies = cookies
        self.num_of_processes = num_of_processes

    def get_dataframes(self, drop_agg_suffix=True):
        "Return a dictionary of dataframes indexed by label - one per data spec.\n\n        Args:\n            drop_agg_suffix (bool): If a time series has only one aggregate, drop the `|<agg-func>` suffix on\n                                    those column names.\n        Returns:\n            Dict[str, pd.DataFrame]: A label-indexed dictionary of data frames.\n        "
        if len(self.ts_data_specs) == 0:
            return {}
        if self.ts_data_specs is None:
            raise InputError("Data spec does not contain any TimeSeriesDataSpecs")
        dataframes = {}
        for tsds in self.ts_data_specs:
            df = self.get_dataframe(tsds.label, drop_agg_suffix=drop_agg_suffix)
            dataframes[tsds.label] = df
        return dataframes

    def get_dataframe(self, label="default", drop_agg_suffix=True):
        if self.ts_data_specs is None:
            raise InputError("Data spec does not contain any TimeSeriesDataSpecs")
        tsds = None
        for ts_data_spec in self.ts_data_specs:
            if ts_data_spec.label == label:
                tsds = ts_data_spec
        if tsds:
            ts_list = []
            ts_res = time_series_v06.get_multiple_time_series_by_id(
                ids=list(set([ts.id for ts in tsds.time_series])), api_key=self.api_key, project=self.project
            )
            id_to_name = {ts["id"]: ts["name"] for ts in ts_res.to_json()}
            for ts in tsds.time_series:
                ts._name = id_to_name[ts.id]
                if not ts.label:
                    ts.label = id_to_name[ts.id]
                ts_dict = dict(
                    name=id_to_name[ts.id],
                    aggregates=(ts.aggregates or tsds.aggregates),
                    missingDataStrategy=ts.missing_data_strategy,
                )
                ts_list.append(ts_dict)
            df = time_series_v05.get_datapoints_frame(
                ts_list,
                tsds.aggregates,
                tsds.granularity,
                tsds.start,
                tsds.end,
                api_key=self.api_key,
                project=self.project,
                cookies=self.cookies,
                processes=self.num_of_processes,
            )
            df = self.__apply_missing_data_strategies(df, ts_list, tsds.missing_data_strategy)
            df = self.__convert_ts_names_to_labels(df, tsds, drop_agg_suffix)
            return df
        return None

    def get_file(self, name):
        "Return files by name as specified in the DataSpec\n\n        Args:\n            name (str): Name of file\n        "
        if not self.files_data_spec:
            raise InputError("Data spec does not contain a FilesDataSpec")
        if isinstance(self.files_data_spec, FilesDataSpec):
            id = self.files_data_spec.file_ids.get(name)
            if id:
                file_bytes = files.download_file(id, get_contents=True, api_key=self.api_key, project=self.project)
                return BytesIO(file_bytes)
            raise InputError("Invalid name")

    def __convert_ts_names_to_labels(self, df, tsds, drop_agg_suffix):
        name_to_label = {}
        for ts in tsds.time_series:
            for agg in ts.aggregates or tsds.aggregates:
                agg_return_name = get_aggregate_func_return_name(agg)
                if (len((ts.aggregates or tsds.aggregates)) > 1) or (not drop_agg_suffix):
                    name_to_label[((ts._name + "|") + agg_return_name)] = (ts.label + "|") + agg_return_name
                else:
                    name_to_label[((ts._name + "|") + agg_return_name)] = ts.label
        return df.rename(columns=name_to_label)

    def __apply_missing_data_strategies(self, df, ts_list, global_missing_data_strategy):
        "Applies missing data strategies to dataframe.\n\n        Local strategies have precedence over global strategy.\n        "
        new_df = df["timestamp"]
        for ts in ts_list:
            name = ts["name"]
            colnames = [colname for colname in df.columns.values if colname.startswith(name)]
            missing_data_strategy = ts.get("missingDataStrategy", global_missing_data_strategy)
            partial_df = df[colnames]
            if missing_data_strategy == "ffill":
                partial_df = df[colnames].fillna(method="pad")
            elif missing_data_strategy and missing_data_strategy.endswith("Interpolation"):
                method = missing_data_strategy[:(-13)].lower()
                partial_df = df[colnames].interpolate(method=method, axis=0)
            new_df = pd.concat([new_df, partial_df], axis=1)
            new_df = new_df.iloc[:, (~new_df.columns.duplicated())]
        return new_df



"Files Module\n\nThis module mirrors the Files API. It allows you to manage files in GCP.\n\nhttps://doc.cognitedata.com/0.5/#Cognite-API-Files\n"
import os
import warnings
import requests
import cognite._utils as _utils
import cognite.config as config
from cognite.v05.dto import FileInfoResponse, FileListResponse


def upload_file(file_name, file_path=None, directory=None, source=None, file_type=None, content_type=None, **kwargs):
    "Upload metadata about a file and get an upload link.\n\n    The link will expire after 30 seconds if not resumable. A resumable upload link is default. Such a link is one-time\n    use and expires after one week. For more information, check this link:\n    https://cloud.google.com/storage/docs/json_api/v1/how-tos/resumable-upload. Use PUT request to upload file with the\n    link returned.\n\n    If file_path is specified, the file will be uploaded directly by the SDK.\n\n    Args:\n        file_name (str):      File name. Max length is 256.\n\n        file_path (str, optional):     Path of file to upload, if omitted a upload link will be returned.\n\n        content_type (str, optional):   MIME type of your file. Required if file_path is specified.\n\n        directory (str, optional):      Directory containing the file. Max length is 512.\n\n        source (str, optional):         Source that this file comes from. Max length is 256.\n\n        file_type (str, optional):      File type. E.g. pdf, css, spreadsheet, .. Max length is 64.\n\n    Keyword Args:\n        api_key (str, optional):        Your api-key.\n\n        project (str, optional):        Project name.\n\n        metadata (dict):      Customized data about the file.\n\n        asset_ids (list):       IDs of assets related to this file.\n\n        resumable (bool):     Whether to generate a resumable URL or not. Default is true.\n\n        overwrite (bool):     Whether to overwrite existing data if duplicate or not. Default is false.\n\n    Returns:\n        dict: A dictionary containing the field fileId and optionally also uploadURL if file_path is omitted.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/files/initupload".format(project)
    headers = {
        "api-key": api_key,
        "content-type": "application/json",
        "accept": "application/json",
        "X-Upload-Content-Type": content_type,
    }
    params = {"resumable": kwargs.get("resumable", True), "overwrite": kwargs.get("overwrite", False)}
    body = {
        "fileName": file_name,
        "directory": directory,
        "source": source,
        "fileType": file_type,
        "metadata": kwargs.get("metadata", None),
        "assetIds": kwargs.get("asset_ids", None),
    }
    res_storage = _utils.post_request(url=url, body=body, headers=headers, params=params, cookies=config.get_cookies())
    result = res_storage.json()["data"]
    if file_path:
        if not content_type:
            warning = "content_type should be specified when directly uploading the file."
            warnings.warn(warning)
        headers = {"content-length": str(os.path.getsize(file_path))}
        with open(file_path, "rb") as file:
            requests.put(result["uploadURL"], data=file, headers=headers)
        result.pop("uploadURL")
    return result


def download_file(id, get_contents=False, **kwargs):
    "Get list of files matching query.\n\n    Args:\n        id (int):                           Path to file to upload, if omitted a upload link will be returned.\n\n        get_contents (bool, optional):      Boolean to determince whether or not to return file contents as string.\n                                            Default is False and download url is returned.\n\n    Keyword Args:\n        api_key (str, optional):            Your api-key.\n\n        project (str, optional):            Project name.\n\n    Returns:\n        bytes: Download link if get_contents is False else file contents.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/files/{}/downloadlink".format(project, id)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = _utils.get_request(url=url, headers=headers, cookies=config.get_cookies())
    if get_contents:
        dl_link = res.json()["data"]
        res = requests.get(dl_link)
        return res.content
    return res.json()["data"]


def delete_files(file_ids, **kwargs):
    "Delete\n\n    Args:\n        file_ids (list[int]):   List of IDs of files to delete.\n\n    Keyword Args:\n        api_key (str):          Your api key.\n\n        project (str):          Your project.\n\n    Returns:\n        List of files deleted and files that failed to delete.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/files/delete".format(project)
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    body = {"items": file_ids}
    res = _utils.post_request(url, body=body, headers=headers)
    return res.json()["data"]


def list_files(name=None, directory=None, file_type=None, source=None, **kwargs):
    "Get list of files matching query.\n\n    Args:\n        name (str, optional):      List all files with this name.\n\n        directory (str, optional):      Directory to list files from.\n\n        source (str, optional):         List files coming from this source.\n\n        file_type (str, optional):      Type of files to list.\n\n    Keyword Args:\n        api_key (str, optional):        Your api-key.\n\n        project (str, optional):        Project name.\n\n        asset_id (list):                Returns all files associated with this asset id.\n\n        sort (str):                     Sort descending or ascending. 'ASC' or 'DESC'.\n\n        limit (int):                    Number of results to return.\n\n        is_uploaded (bool):             List only uploaded files if true. If false, list only other files. If not set,\n                                        list all files without considering whether they are uploaded or not.\n\n        autopaging (bool):              Whether or not to automatically page through results. If set to true, limit will\n                                        be disregarded. Defaults to False.\n\n        cursor (str):                   Cursor to use for paging through results.\n\n    Returns:\n        v05.dto.FileListResponse: A data object containing the requested files information.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/files".format(project)
    headers = {"api-key": api_key, "accept": "application/json"}
    params = {
        "assetId": kwargs.get("asset_id"),
        "dir": directory,
        "name": name,
        "type": file_type,
        "source": source,
        "isUploaded": kwargs.get("is_uploaded"),
        "sort": kwargs.get("sort"),
        "limit": (kwargs.get("limit", 100) if (not kwargs.get("autopaging")) else 10000),
        "cursor": kwargs.get("cursor"),
    }
    file_list = []
    res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
    file_list.extend(res.json()["data"]["items"])
    next_cursor = res.json()["data"].get("nextCursor", None)
    while next_cursor and kwargs.get("autopaging"):
        params["cursor"] = next_cursor
        res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
        file_list.extend(res.json()["data"]["items"])
        next_cursor = res.json()["data"].get("nextCursor", None)
    return FileListResponse(
        {
            "data": {
                "nextCursor": next_cursor,
                "previousCursor": res.json()["data"].get("previousCursor"),
                "items": file_list,
            }
        }
    )


def get_file_info(id, **kwargs):
    "Returns information about a file.\n\n    Args:\n        id (int):                   Id of the file.\n\n    Keyword Args:\n        api_key (str, optional):    Your api-key.\n\n        project (str, optional):    Project name.\n\n    Returns:\n        v05.dto.FileInfoResponse: A data object containing the requested file information.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/files/{}".format(project, id)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = _utils.get_request(url, headers=headers)
    return FileInfoResponse(res.json())



"Project Configuration Module.\n\nThis module allows you to set an api-key and a project for your python project.\n"
import os
from cognite_logger import cognite_logger
from cognite._constants import BASE_URL, RETRY_LIMIT

_CONFIG_API_KEY = os.getenv("COGNITE_API_KEY", "")
_CONFIG_PROJECT = os.getenv("COGNITE_PROJECT", "")
_CONFIG_COOKIES = {}
_CONFIG_BASE_URL = os.getenv("COGNITE_BASE_URL")
_CONFIG_RETRIES = os.getenv("COGNITE_NUM_RETRIES")


def configure_session(api_key="", project="", cookies=None, debug=False):
    "Sets session variables.\n\n    Args:\n        api_key (str):  Api-key for current project.\n\n        cookies (dict): Cookies to pass with requests.\n\n        project (str):  Project name for current project.\n\n        debug (bool): Whether or not to ouptut a debug log.\n    "
    global _CONFIG_API_KEY, _CONFIG_COOKIES, _CONFIG_PROJECT
    _CONFIG_API_KEY = api_key
    _CONFIG_PROJECT = project
    _CONFIG_COOKIES = {} if (cookies is None) else cookies
    cognite_logger.configure_logger(logger_name="cognite-sdk", log_json=True, log_level=("DEBUG" if debug else "INFO"))


def get_config_variables(api_key, project):
    "Returns current project config variables unless other is specified.\n\n    Args:\n        api_key (str):  Other specified api-key.\n\n        project (str):  Other specified project name.\n\n    Returns:\n        tuple: api-key and project name belonging to current project unless other is specified.\n    "
    if api_key is None:
        api_key = _CONFIG_API_KEY
    if project is None:
        project = _CONFIG_PROJECT
    return (api_key, project)


def get_cookies(cookies=None):
    "Returns cookies set for the current session.\n\n    Returns:\n        dict: Cookies for current session.\n    "
    if cookies is None:
        cookies = _CONFIG_COOKIES
    return cookies


def set_base_url(url=None):
    "Sets the base url for requests made from the SDK.\n\n    Args:\n        url (str):  URL to set. Set this to None to use default url.\n    "
    global _CONFIG_BASE_URL
    _CONFIG_BASE_URL = url


def get_base_url():
    "Returns the current base url for requests made from the SDK.\n    Args:\n        api_version (float): Version of API to use for base_url\n    Returns:\n        str: current base url.\n    "
    return _CONFIG_BASE_URL or BASE_URL


def set_number_of_retries(retries=None):
    "Sets the number of retries attempted for requests made from the SDK.\n\n    Args:\n        retries (int):  Number of retries to attempt. Set this to None to use default num of retries.\n    "
    global _CONFIG_RETRIES
    _CONFIG_RETRIES = retries


def get_number_of_retries():
    "Returns the current number of retries attempted for requests made from the SDK.\n\n    Returns:\n        int: current number of retries attempted for requests.\n    "
    return RETRY_LIMIT if (_CONFIG_RETRIES is None) else _CONFIG_RETRIES



"Constants\n\nThis module contains constants used in the Cognite Python SDK.\n\nThis module is protected and should not used by end-users.\n\nAttributes:\n    BASE_URL (str):    Base url for Cognite API. Should have correct version number.\n    LIMIT (int):       Limit on how many datapoints should be returned from the API when fetching data using the\n                        timeseries module.\n    RETRY_LIMIT (int): Number of retries to perform if a request to the API should fail.\n"
import os

BASE_URL = "https://api.cognitedata.com"
LIMIT = 100000
LIMIT_AGG = 10000
RETRY_LIMIT = 0
NUM_OF_WORKERS = 10
TIMESERIES = "timeSeries"
AGGREGATES = "aggregates"
GRANULARITY = "granularity"
START = "start"
END = "end"
MISSING_DATA_STRATEGY = "missingDataStrategy"
LABEL = "label"



"Datapoints Module\n\nThis module mirrors the Datapoints API. It allows you to fetch data from the api and output it in various formats.\n\nhttps://doc.cognitedata.com/0.6/#Cognite-API-Datapoints\n"
from concurrent.futures import ThreadPoolExecutor as Pool
from functools import partial
import cognite._constants as _constants
import cognite._utils as _utils
import cognite.config as config
from cognite.v05.dto import DatapointsResponse


def get_datapoints(id, start, end=None, aggregates=None, granularity=None, **kwargs):
    "Returns a DatapointsObject containing a list of datapoints for the given query.\n\n    This method will automate paging for the user and return all data for the given time period.\n\n    Args:\n        id (int):             The unique id of the timeseries to retrieve data for.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions\n                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n    Keyword Arguments:\n        include_outside_points (bool):      No description\n\n        processes (int):        Number of download processes to run in parallell. Defaults to number returned by cpu_count().\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n        limit (str):            Max number of datapoints to return. If limit is specified, this method will not automate\n                                paging and will return a maximum of 100,000 dps.\n\n    Returns:\n        v05.dto.DatapointsResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    (start, end) = _utils.interval_to_ms(start, end)
    if kwargs.get("limit"):
        return _get_datapoints_user_defined_limit(
            id,
            aggregates,
            granularity,
            start,
            end,
            limit=kwargs.get("limit"),
            include_outside_points=kwargs.get("include_outside_points", False),
            api_key=api_key,
            project=project,
        )
    diff = end - start
    num_of_processes = kwargs.get("processes", _constants.NUM_OF_WORKERS)
    granularity_ms = 1
    if granularity:
        granularity_ms = _utils.granularity_to_ms(granularity)
    steps = min(num_of_processes, max(1, int((diff / granularity_ms))))
    step_size = _utils.round_to_nearest(int((diff / steps)), base=granularity_ms)
    step_starts = [(start + (i * step_size)) for i in range(steps)]
    args = [{"start": start, "end": (start + step_size)} for start in step_starts]
    partial_get_dps = partial(
        _get_datapoints_helper_wrapper,
        id=id,
        aggregates=aggregates,
        granularity=granularity,
        include_outside_points=kwargs.get("include_outside_points", False),
        api_key=api_key,
        project=project,
    )
    with Pool(steps) as p:
        datapoints = p.map(partial_get_dps, args)
    concat_dps = []
    [concat_dps.extend(el) for el in datapoints]
    return DatapointsResponse({"data": {"items": [{"id": id, "datapoints": concat_dps}]}})


def _get_datapoints_helper_wrapper(args, id, aggregates, granularity, include_outside_points, api_key, project):
    return _get_datapoints_helper(
        id,
        aggregates,
        granularity,
        args["start"],
        args["end"],
        api_key=api_key,
        project=project,
        include_outside_points=include_outside_points,
    )


def _get_datapoints_helper(id, aggregates=None, granularity=None, start=None, end=None, **kwargs):
    "Returns a list of datapoints for the given query.\n\n    This method will automate paging for the given time period.\n\n    Args:\n        id (int):       The unique id of the timeseries to retrieve data for.\n\n        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions\n                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n    Keyword Arguments:\n        include_outside_points (bool):  No description.\n\n        api_key (str):          Your api-key. Obligatory in this helper method.\n\n        project (str):          Project name. Obligatory in this helper method.\n\n    Returns:\n        list of datapoints: A list containing datapoint dicts.\n    "
    (api_key, project) = (kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/timeseries/{}/data".format(project, id)
    limit = _constants.LIMIT if (aggregates is None) else _constants.LIMIT_AGG
    params = {
        "aggregates": aggregates,
        "granularity": granularity,
        "limit": limit,
        "start": start,
        "end": end,
        "includeOutsidePoints": kwargs.get("include_outside_points", False),
    }
    headers = {"api-key": api_key, "accept": "application/json"}
    datapoints = []
    while ((not datapoints) or (len(datapoints[(-1)]) == limit)) and (params["end"] > params["start"]):
        res = _utils.get_request(url, params=params, headers=headers)
        res = res.json()["data"]["items"][0]["datapoints"]
        if not res:
            break
        datapoints.append(res)
        latest_timestamp = int(datapoints[(-1)][(-1)]["timestamp"])
        params["start"] = latest_timestamp + (_utils.granularity_to_ms(granularity) if granularity else 1)
    dps = []
    [dps.extend(el) for el in datapoints]
    return dps


def _get_datapoints_user_defined_limit(id, aggregates, granularity, start, end, limit, **kwargs):
    "Returns a DatapointsResponse object with the requested data.\n\n    No paging or parallelizing is done.\n\n    Args:\n        id (int):       The unique id of the timeseries to retrieve data for.\n\n        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions\n                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n        limit (str):            Max number of datapoints to return. Max is 100,000.\n\n    Keyword Arguments:\n        include_outside_points (bool):  No description.\n\n        api_key (str):          Your api-key. Obligatory in this helper method.\n\n        project (str):          Project name. Obligatory in this helper method.\n    Returns:\n        v05.dto.DatapointsResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = (kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/timeseries/{}/data".format(project, id)
    params = {
        "aggregates": aggregates,
        "granularity": granularity,
        "limit": limit,
        "start": start,
        "end": end,
        "includeOutsidePoints": kwargs.get("include_outside_points", False),
    }
    headers = {"api-key": api_key, "accept": "application/json"}
    res = _utils.get_request(url, params=params, headers=headers)
    res = res.json()["data"]["items"][0]["datapoints"]
    return DatapointsResponse({"data": {"items": [{"id": id, "datapoints": res}]}})



"Timeseries Module\n\nThis module mirrors the Timeseries API. It allows you to fetch data from the api and output it in various formats.\n\nhttps://doc.cognitedata.com/0.4/#Cognite-API-Time-series\n"
import io
import os
import warnings
from concurrent.futures import ThreadPoolExecutor as Pool
from functools import partial
from urllib.parse import quote
import pandas as pd
import cognite._constants as _constants
import cognite._utils as _utils
import cognite.config as config
from cognite.auxiliary._protobuf_descriptors import _api_timeseries_data_v1_pb2
from cognite.v04.dto import (
    Datapoint,
    DatapointsResponse,
    DatapointsResponseIterator,
    LatestDatapointResponse,
    TimeSeries,
    TimeSeriesResponse,
    TimeseriesWithDatapoints,
)


def get_datapoints(tag_id, aggregates=None, granularity=None, start=None, end=None, **kwargs):
    "Returns a DatapointsObject containing a list of datapoints for the given query.\n\n    This method will automate paging for the user and return all data for the given time period.\n\n    Args:\n        tag_id (str):           The tag_id to retrieve data for.\n\n        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions\n                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n    Keyword Arguments:\n        protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.\n                                Defaults to True.\n\n        processes (int):        Number of download processes to run in parallell. Defaults to number returned by cpu_count().\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v04.dto.DatapointsResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    (start, end) = _utils.interval_to_ms(start, end)
    diff = end - start
    num_of_processes = kwargs.get("processes", _constants.NUM_OF_WORKERS)
    granularity_ms = 1
    if granularity:
        granularity_ms = _utils.granularity_to_ms(granularity)
    steps = min(num_of_processes, max(1, int((diff / granularity_ms))))
    step_size = _utils.round_to_nearest(int((diff / steps)), base=granularity_ms)
    step_starts = [(start + (i * step_size)) for i in range(steps)]
    args = [{"start": start, "end": (start + step_size)} for start in step_starts]
    partial_get_dps = partial(
        _get_datapoints_helper_wrapper,
        tag_id=tag_id,
        aggregates=aggregates,
        granularity=granularity,
        protobuf=kwargs.get("protobuf", True),
        api_key=api_key,
        project=project,
    )
    if steps == 1:
        dps = _get_datapoints_helper(
            tag_id,
            aggregates,
            granularity,
            start,
            end,
            protobuf=kwargs.get("protobuf", True),
            api_key=api_key,
            project=project,
        )
        return DatapointsResponse({"data": {"items": [{"tagId": tag_id, "datapoints": dps}]}})
    with Pool(steps) as p:
        datapoints = p.map(partial_get_dps, args)
    concat_dps = []
    [concat_dps.extend(el) for el in datapoints]
    return DatapointsResponse({"data": {"items": [{"tagId": tag_id, "datapoints": concat_dps}]}})


def _get_datapoints_helper_wrapper(args, tag_id, aggregates, granularity, protobuf, api_key, project):
    return _get_datapoints_helper(
        tag_id, aggregates, granularity, args["start"], args["end"], protobuf=protobuf, api_key=api_key, project=project
    )


def _get_datapoints_helper(tag_id, aggregates=None, granularity=None, start=None, end=None, **kwargs):
    "Returns a list of datapoints for the given query.\n\n    This method will automate paging for the given time period.\n\n    Args:\n        tag_id (str):           The tag_id to retrieve data for.\n\n        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions\n                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n    Keyword Arguments:\n        protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.\n                                Defaults to True.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        list of datapoints: A list containing datapoint dicts.\n    "
    (api_key, project) = (kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/timeseries/data/{}".format(project, quote(tag_id, safe=""))
    use_protobuf = kwargs.get("protobuf", True) and (aggregates is None)
    limit = _constants.LIMIT if (aggregates is None) else _constants.LIMIT_AGG
    params = {"aggregates": aggregates, "granularity": granularity, "limit": limit, "start": start, "end": end}
    headers = {"api-key": api_key, "accept": ("application/protobuf" if use_protobuf else "application/json")}
    datapoints = []
    while ((not datapoints) or (len(datapoints[(-1)]) == limit)) and (params["end"] > params["start"]):
        res = _utils.get_request(url, params=params, headers=headers)
        if use_protobuf:
            ts_data = _api_timeseries_data_v1_pb2.TimeseriesData()
            ts_data.ParseFromString(res.content)
            res = [{"timestamp": p.timestamp, "value": p.value} for p in ts_data.numericData.points]
        else:
            res = res.json()["data"]["items"][0]["datapoints"]
        if not res:
            warning = "An interval with no data has been requested ({}, {}).".format(params["start"], params["end"])
            warnings.warn(warning)
            break
        datapoints.append(res)
        latest_timestamp = int(datapoints[(-1)][(-1)]["timestamp"])
        params["start"] = latest_timestamp + (_utils.granularity_to_ms(granularity) if granularity else 1)
    dps = []
    [dps.extend(el) for el in datapoints]
    return dps


def _split_TimeseriesWithDatapoints_if_over_limit(timeseries_with_datapoints, limit):
    "Takes a TimeseriesWithDatapoints and splits it into multiple so that each has a max number of datapoints equal\n    to the limit given.\n\n    Args:\n        timeseries_with_datapoints (TimeseriesWithDatapoints): The timeseries with data to potentially split up.\n\n    Returns:\n        A list of TimeSeriesWithDatapoints where each has a maximum number of datapoints equal to the limit given.\n    "
    timeseries_with_datapoints_list = []
    if len(timeseries_with_datapoints.datapoints) > limit:
        i = 0
        while i < len(timeseries_with_datapoints.datapoints):
            timeseries_with_datapoints_list.append(
                TimeseriesWithDatapoints(
                    tagId=timeseries_with_datapoints.tagId,
                    datapoints=timeseries_with_datapoints.datapoints[i : (i + limit)],
                )
            )
            i += limit
    else:
        timeseries_with_datapoints_list.append(timeseries_with_datapoints)
    return timeseries_with_datapoints_list


def post_multi_tag_datapoints(timeseries_with_datapoints, **kwargs):
    "Insert data into multiple timeseries.\n\n    Args:\n        timeseries_with_datapoints (List[v04.dto.TimeseriesWithDatapoints]): The timeseries with data to insert.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/timeseries/data".format(project)
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    ul_dps_limit = 100000
    timeseries_with_datapoints_limited = []
    for entry in timeseries_with_datapoints:
        timeseries_with_datapoints_limited.extend(_split_TimeseriesWithDatapoints_if_over_limit(entry, ul_dps_limit))
    timeseries_to_upload_binned = _utils.first_fit(
        list_items=timeseries_with_datapoints_limited, max_size=ul_dps_limit, get_count=(lambda x: len(x.datapoints))
    )
    for bin in timeseries_to_upload_binned:
        body = {
            "items": [
                {"tagId": ts_with_data.tagId, "datapoints": [dp.__dict__ for dp in ts_with_data.datapoints]}
                for ts_with_data in bin
            ]
        }
        res = _utils.post_request(url, body=body, headers=headers)
    return res.json()


def post_datapoints(tag_id, datapoints, **kwargs):
    "Insert a list of datapoints.\n\n    Args:\n        tag_id (str):       ID of timeseries to insert to.\n\n        datapoints (list[v04.dto.Datapoint): List of datapoint data transfer objects to insert.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/timeseries/data/{}".format(project, quote(tag_id, safe=""))
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    ul_dps_limit = 100000
    i = 0
    while i < len(datapoints):
        body = {"items": [dp.__dict__ for dp in datapoints[i : (i + ul_dps_limit)]]}
        res = _utils.post_request(url, body=body, headers=headers)
        i += ul_dps_limit
    return res.json()


def get_latest(tag_id, **kwargs):
    "Returns a LatestDatapointObject containing the latest datapoint for the given tag_id.\n\n    Args:\n        tag_id (str):           The tag_id to retrieve data for.\n\n    Keyword Arguments:\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v04.dto.LatestDatapointsResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/timeseries/latest/{}".format(project, quote(tag_id, safe=""))
    headers = {"api-key": api_key, "accept": "application/json"}
    res = _utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return LatestDatapointResponse(res.json())


def get_multi_tag_datapoints(datapoints_queries, aggregates=None, granularity=None, start=None, end=None, **kwargs):
    "Returns a list of DatapointsObjects each of which contains a list of datapoints for the given timeseries.\n\n    This method will automate paging for the user and return all data for the given time period(s).\n\n    Args:\n        datapoints_queries (list[v04.dto.DatapointsQuery]): The list of DatapointsQuery objects specifying which\n                                                                    timeseries to retrieve data for.\n\n        aggregates (list, optional):    The list of aggregate functions you wish to apply to the data. Valid aggregate\n                                        functions are: 'average/avg, max, min, count, sum, interpolation/int,\n                                        stepinterpolation/step'.\n\n        granularity (str):              The granularity of the aggregate values. Valid entries are : 'day/d, hour/h,\n                                        minute/m, second/s', or a multiple of these indicated by a number as a prefix\n                                        e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n    Keyword Arguments:\n        api_key (str):                  Your api-key.\n\n        project (str):                  Project name.\n\n    Returns:\n        list(v04.dto.DatapointsResponse): A list of data objects containing the requested data with several getter methods\n        with different output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/timeseries/dataquery".format(project)
    (start, end) = _utils.interval_to_ms(start, end)
    num_of_dpqs_with_agg = 0
    num_of_dpqs_raw = 0
    for dpq in datapoints_queries:
        if ((dpq.aggregateFunctions is None) and (aggregates is None)) or (dpq.aggregateFunctions == ""):
            num_of_dpqs_raw += 1
        else:
            num_of_dpqs_with_agg += 1
    items = []
    for dpq in datapoints_queries:
        if (dpq.aggregateFunctions is None) and (aggregates is None):
            dpq.limit = int((_constants.LIMIT / num_of_dpqs_raw))
        else:
            dpq.limit = int((_constants.LIMIT_AGG / num_of_dpqs_with_agg))
        items.append(dpq.__dict__)
    body = {
        "items": items,
        "aggregateFunctions": (",".join(aggregates) if (aggregates is not None) else None),
        "granularity": granularity,
        "start": start,
        "end": end,
    }
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    datapoints_responses = []
    has_incomplete_requests = True
    while has_incomplete_requests:
        res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies()).json()["data"][
            "items"
        ]
        datapoints_responses.append(res)
        has_incomplete_requests = False
        for (i, dpr) in enumerate(res):
            dpq = datapoints_queries[i]
            if len(dpr["datapoints"]) == dpq.limit:
                has_incomplete_requests = True
                latest_timestamp = dpr["datapoints"][(-1)]["timestamp"]
                ts_granularity = granularity if (dpq.granularity is None) else dpq.granularity
                next_start = latest_timestamp + (_utils.granularity_to_ms(ts_granularity) if ts_granularity else 1)
            else:
                next_start = end - 1
                if datapoints_queries[i].end:
                    next_start = datapoints_queries[i].end - 1
            datapoints_queries[i].start = next_start
    results = [{"data": {"items": [{"tagId": dpq.tagId, "datapoints": []}]}} for dpq in datapoints_queries]
    for res in datapoints_responses:
        for (i, ts) in enumerate(res):
            results[i]["data"]["items"][0]["datapoints"].extend(ts["datapoints"])
    return DatapointsResponseIterator([DatapointsResponse(result) for result in results])


def get_datapoints_frame(tag_ids, aggregates, granularity, start=None, end=None, **kwargs):
    "Returns a pandas dataframe of datapoints for the given tag_ids all on the same timestamps.\n\n    This method will automate paging for the user and return all data for the given time period.\n\n    Args:\n        tag_ids (list):     The list of tag_ids to retrieve data for. Each tag_id can be either a string containing the\n                            tag_id or a dictionary containing the tag_id and a list of specific aggregate functions.\n\n        aggregates (list):  The list of aggregate functions you wish to apply to the data for which you have not\n                            specified an aggregate function. Valid aggregate functions are: 'average/avg, max, min,\n                            count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):  The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                            second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n    Keyword Arguments:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n        processes (int):    Number of download processes to run in parallell. Defaults to number returned by cpu_count().\n\n    Returns:\n        pandas.DataFrame: A pandas dataframe containing the datapoints for the given tag_ids. The datapoints for all the\n        tag_ids will all be on the same timestamps.\n\n    Note:\n        The ``tag_ids`` parameter can take a list of strings and/or dicts on the following formats::\n\n            Using strings:\n                ['<tag_id1>', '<tag_id2>']\n\n            Using dicts:\n                [{'tagId': '<tag_id1>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']},\n                {'tagId': '<tag_id2>', 'aggregates': []}]\n\n            Using both:\n                ['<tagid1>', {'tagId': '<tag_id2>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']}]\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    (start, end) = _utils.interval_to_ms(start, end)
    diff = end - start
    num_of_processes = kwargs.get("processes", _constants.NUM_OF_WORKERS)
    granularity_ms = 1
    if granularity:
        granularity_ms = _utils.granularity_to_ms(granularity)
    steps = min(num_of_processes, max(1, int((diff / granularity_ms))))
    step_size = _utils.round_to_nearest(int((diff / steps)), base=granularity_ms)
    step_starts = [(start + (i * step_size)) for i in range(steps)]
    args = [{"start": start, "end": (start + step_size)} for start in step_starts]
    partial_get_dpsf = partial(
        _get_datapoints_frame_helper_wrapper,
        tag_ids=tag_ids,
        aggregates=aggregates,
        granularity=granularity,
        api_key=api_key,
        project=project,
    )
    if steps == 1:
        return _get_datapoints_frame_helper(
            tag_ids, aggregates, granularity, start, end, api_key=api_key, project=project
        )
    with Pool(steps) as p:
        dataframes = p.map(partial_get_dpsf, args)
    df = pd.concat(dataframes).drop_duplicates(subset="timestamp").reset_index(drop=True)
    return df


def _get_datapoints_frame_helper_wrapper(args, tag_ids, aggregates, granularity, api_key, project):
    return _get_datapoints_frame_helper(
        tag_ids, aggregates, granularity, args["start"], args["end"], api_key=api_key, project=project
    )


def _get_datapoints_frame_helper(tag_ids, aggregates, granularity, start=None, end=None, **kwargs):
    "Returns a pandas dataframe of datapoints for the given tag_ids all on the same timestamps.\n\n    This method will automate paging for the user and return all data for the given time period.\n\n    Args:\n        tag_ids (list):     The list of tag_ids to retrieve data for. Each tag_id can be either a string containing the\n                            tag_id or a dictionary containing the tag_id and a list of specific aggregate functions.\n\n        aggregates (list):  The list of aggregate functions you wish to apply to the data for which you have not\n                            specified an aggregate function. Valid aggregate functions are: 'average/avg, max, min,\n                            count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):  The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                            second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n    Keyword Arguments:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        pandas.DataFrame: A pandas dataframe containing the datapoints for the given tag_ids. The datapoints for all the\n        tag_ids will all be on the same timestamps.\n\n    Note:\n        The ``tag_ids`` parameter can take a list of strings and/or dicts on the following formats::\n\n            Using strings:\n                ['<tag_id1>', '<tag_id2>']\n\n            Using dicts:\n                [{'tagId': '<tag_id1>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']},\n                {'tagId': '<tag_id2>', 'aggregates': []}]\n\n            Using both:\n                ['<tagid1>', {'tagId': '<tag_id2>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']}]\n    "
    (api_key, project) = (kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/timeseries/dataframe".format(project)
    num_aggregates = 0
    for tag in tag_ids:
        if isinstance(tag, str) or (tag.get("aggregates") is None):
            num_aggregates += len(aggregates)
        else:
            num_aggregates += len(tag["aggregates"])
    per_tag_limit = int((_constants.LIMIT / num_aggregates))
    body = {
        "items": [
            (
                {"tagId": "{}".format(tag_id)}
                if isinstance(tag_id, str)
                else {"tagId": "{}".format(tag_id["tagId"]), "aggregates": tag_id.get("aggregates", [])}
            )
            for tag_id in tag_ids
        ],
        "aggregates": aggregates,
        "granularity": granularity,
        "start": start,
        "end": end,
        "limit": per_tag_limit,
    }
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "text/csv"}
    dataframes = []
    while ((not dataframes) or (dataframes[(-1)].shape[0] == per_tag_limit)) and (body["end"] > body["start"]):
        res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
        dataframes.append(
            pd.read_csv(io.StringIO(res.content.decode((res.encoding if res.encoding else res.apparent_encoding))))
        )
        if dataframes[(-1)].empty:
            warning = "An interval with no data has been requested ({}, {}).".format(body["start"], body["end"])
            warnings.warn(warning)
            break
        latest_timestamp = int(dataframes[(-1)].iloc[((-1), 0)])
        body["start"] = latest_timestamp + _utils.granularity_to_ms(granularity)
    return pd.concat(dataframes).reset_index(drop=True)


def get_timeseries(prefix=None, description=None, include_metadata=False, asset_id=None, path=None, **kwargs):
    "Returns a TimeseriesObject containing the requested timeseries.\n\n    Args:\n        prefix (str):           List timeseries with this prefix in the name.\n\n        description (str):      Filter timeseries taht contains this string in its description.\n\n        include_metadata (bool):    Decide if the metadata field should be returned or not. Defaults to False.\n\n        asset_id (int):        Get timeseries related to this asset.\n\n        path (str):             Get timeseries under this asset path branch.\n\n    Keyword Arguments:\n        limit (int):            Number of results to return.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n        autopaging (bool):      Whether or not to automatically page through results. If set to true, limit will be\n                                disregarded. Defaults to False.\n\n    Returns:\n        v04.dto.TimeSeriesResponse: A data object containing the requested timeseries with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/timeseries".format(project)
    headers = {"api-key": api_key, "accept": "application/json"}
    params = {
        "q": prefix,
        "description": description,
        "includeMetadata": include_metadata,
        "assetId": asset_id,
        "path": path,
        "limit": (kwargs.get("limit", 10000) if (not kwargs.get("autopaging")) else 10000),
    }
    timeseries = []
    res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
    timeseries.extend(res.json()["data"]["items"])
    next_cursor = res.json()["data"].get("nextCursor")
    while next_cursor and kwargs.get("autopaging"):
        params["cursor"] = next_cursor
        res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
        timeseries.extend(res.json()["data"]["items"])
        next_cursor = res.json()["data"].get("nextCursor")
    return TimeSeriesResponse(
        {
            "data": {
                "nextCursor": next_cursor,
                "previousCursor": res.json()["data"].get("previousCursor"),
                "items": timeseries,
            }
        }
    )


def post_time_series(time_series, **kwargs):
    "Create a new time series.\n\n    Args:\n        timeseries (list[v04.dto.TimeSeries]):   List of time series data transfer objects to create.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/timeseries".format(project)
    body = {"items": [ts.__dict__ for ts in time_series]}
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    res = _utils.post_request(url, body=body, headers=headers)
    return res.json()


def update_time_series(time_series, **kwargs):
    "Update an existing time series.\n\n    For each field that can be updated, a null value indicates that nothing should be done.\n\n    Args:\n        timeseries (list[v04.dto.TimeSeries]):   List of time series data transfer objects to update.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/timeseries".format(project)
    body = {"items": [ts.__dict__ for ts in time_series]}
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    res = _utils.put_request(url, body=body, headers=headers)
    return res.json()



import pickle

from cognite.data_transfer_service import DataTransferService


class Model:
    def __init__(self, model):
        self._model = model

    @staticmethod
    def train(file_io, data_spec, **kwargs):
        """The method to train your model.
        Should produce at least a serialized model. Can also produce other
        serialized artefacts that you will need when predicting.

        Args:
            file_io:    A callable which allows you to write to model hosting storage.
                        Works just like the built-in function open
        Keyword Args:
            Any user defined arguments

        Returns:
            None
        """
        my_model = {"input": "result"}

        with file_io("yourmodel.pkl", "wb") as f:
            pickle.dump(my_model, f)
        pass

    def predict(self, instance, data_spec, **kwargs):
        """Method to perform predictions on your model.

        Args:
            instance: The input to your model.
        Keyword Args:
            Any user defined arguments
        Returns:
            Json serializable output from your model.
        """
        return

    @staticmethod
    def load(file_io):
        """Method to load your serialzed model into memory

        Can also load other artifacts such as preprocessors
        Args:
            file_io:    A callable which allows you to write to model hosting storage.
                        Works just like the built-in function open
        Returns:
            An instance of Model
        """
        with file_io("yourmodel.pkl", "rb") as f:
            model = pickle.load(f)
        return Model(model)



from setuptools import find_packages, setup

REQUIRED_PACKAGES = []

setup(
    name="my_model",
    version="0.1",
    packages=find_packages(),
    include_package_data=True,
    install_requires=REQUIRED_PACKAGES,
)
